from types import SimpleNamespace

import click


@click.command()
@click.option(
    "-infile",
    required=False,
    default="new_x.npy",
    type=click.File("rb"),
    help=".npy file with data to predict",
)
@click.option(
    "-reg",
    required=False,
    default="random_forest_reg.pkl",
    type=str,
    help="pickled regressor file",
)
@click.option(
    "-outfile",
    required=False,
    default="new_y.npy",
    type=str,
    help="file to store the new predictions",
)
def cli(infile, reg, outfile):
    """
    Use a regressor to make a prediction
    """
    from spellbook.ml import predict

    args = SimpleNamespace(**{"infile": infile, "reg": reg, "outfile": outfile})
    predict.predict(args)



from types import SimpleNamespace

import click


@click.command()
@click.option(
    "-input",
    required=False,
    default="results.json",
    type=str,
    help=".json file with data in it",
)
@click.option(
    "-output",
    required=False,
    default="results.npz",
    type=str,
    help=".npz file with the arrays",
)
@click.option(
    "-schema",
    required=False,
    default="auto",
    type=str,
    help="schema for a single sample that says what data to translate. Defaults to whole first node. Can be a comma-delimited list of subpaths, eg inputs,outputs/scalars,metadata",
)
def cli(input, output, schema):
    """
    Flatten sample json file into numpy", filtering with an external schema.
    """
    from spellbook.data_formatting import translator

    args = SimpleNamespace(**{"input": input, "output": output, "schema": schema})
    translator.process_args(args)



from types import SimpleNamespace

import click


@click.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    required=False,
    default=False,
    type=bool,
    help="Force write the target file",
)
@click.argument(
    "target",
    required=True,
    type=str,
)
@click.argument(
    "source",
    required=True,
    type=str,
    nargs=-1,
)
def cli(force, target, source):
    """
    stacker for npz files.
    """
    from spellbook.data_formatting import stack_npz

    args = SimpleNamespace(**{"force": force, "target": target, "source": source})
    stack_npz.process_args(args)



from __future__ import print_function

import json
import sys
from itertools import zip_longest
from uuid import uuid4

from spellbook.data_formatting.conduit.python import conduit_bundler as cb


WARN = ""
try:
    import conduit
except:
    WARN = "\nWARNING: conduit not found."


def grouper(iterable, n):
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=None)


def savename(file_no, args):
    if args.chunk_size:
        base, ext = args.outfile.split(".")
        name = f"{base}_{file_no:03}.{ext}"
        return name
    return args.outfile


def make_schema_compatible(original_node, add_uuid):
    node = conduit.Node()
    if add_uuid:
        node[str(uuid4())] = original_node
    else:
        node = original_node
    return node


def process_args(args):
    print(WARN)
    files = args.infiles
    nfiles = len(files)
    if not args.chunk_size:
        chunk_size = nfiles
    else:
        chunk_size = args.chunk_size

    fileno = 0
    results = []
    for group in grouper(files, chunk_size):
        result = conduit.Node()
        for path in group:
            if not path:
                continue
            try:
                subnode = cb.load_node(path)
                subnode = make_schema_compatible(subnode, args.add_uuid)
                for top_path in subnode.child_names():

                    if top_path in results:
                        print("Error! Already in results: " + top_path)
                        new_path = "-".join((top_path, str(uuid4())))
                        print("Renaming duplicate to node to: " + new_path)
                    else:
                        new_path = top_path

                    result[new_path] = subnode[top_path]
                    results.append(top_path)
            except:
                print("Unable to load " + path)

        cb.dump_node(result, savename(fileno, args))
        fileno = fileno + 1



import logging
import os
import sys
import traceback

import click

from spellbook import VERSION
from spellbook.log_formatter import setup_logging


LOG = logging.getLogger("spellbook")
PLUGIN_DIR = os.path.join(os.path.dirname(__file__), "commands")


class SpellbookCLI(click.MultiCommand):
    def list_commands(self, ctx):
        """
        Avoids file reads for max speed.
        """
        return [
            "collect",
            "conduit-collect",
            "conduit-translate",
            "learn",
            "make-samples",
            "predict",
            "serialize",
            "stack-npz",
            "translate",
        ]

    def list_commands_dynamically(self, ctx):
        rv = []
        for filename in os.listdir(PLUGIN_DIR):
            if filename.startswith("__"):
                continue
            if filename.endswith(".py"):
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(PLUGIN_DIR, name + ".py")
        if not os.path.isfile(fn):
            return
        with open(fn) as f:
            code = compile(f.read(), fn, "exec")
            eval(code, ns, ns)
        return ns["cli"]


@click.command(cls=SpellbookCLI)
@click.option(
    "--level",
    required=False,
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
    help="set the logger level",
)
@click.version_option(VERSION)
def spellbook(level):
    setup_logging(logger=LOG, log_level=level.upper(), colors=True)


def main():
    if len(sys.argv) == 1:
        with click.Context(spellbook) as ctx:
            click.echo(spellbook.get_help(ctx))
        return 1
    try:
        spellbook()
    except Exception as e:
        # LOG.debug(traceback.format_exc())
        print(traceback.format_exc())
        LOG.error(str(e))
        return 1


if __name__ == "__main__":
    sys.exit(main())



import os

import numpy as np

from spellbook.ml.learn import stack_arrays


def test_stack_arrays():
    data1 = np.random.random((3, 2))
    data2 = np.zeros((3, 2))
    data3 = np.ones((3, 2))
    np.savez("temp.npz", F1=data1, F2=data2, F3=data3)
    loaded = np.load("temp.npz")
    os.remove("temp.npz")
    feature_names = "F1,F2,F3"
    result = stack_arrays(loaded, feature_names)
    assert result.shape == (2, 9)



import ast
import sys

import numpy as np
import pyDOE2 as doe
from scipy.stats.distributions import norm

from spellbook.commands import CliCommand


def scale_samples(samples_norm, limits, limits_norm=(0, 1), do_log=False):
    """Scale samples to new limits, either log10 or linearly.

    Args:
        samples_norm (ndarray): The normalized samples to scale,
            with dimensions (nsamples,ndims).
        limits (list of tuples): A list of (min, max) for the various
            dimensions. Length of list is ndims.
        limits_norm (tuple of floats, optional): The (min, max) from which
            samples_norm were drawn. Defaults to (0,1).
        do_log (boolean or list of booleans, optional): Whether
            to log10 scale each dimension. Either a single boolean or
            a list of length ndims, for each dimension.
            Defaults to ndims*[False].

    Returns:
        ndarray: The scaled samples.

    Note:
        We follow the sklearn convention of requiring samples to be
        given as an (nsamples, ndims) array.

        To transform 1-D arrays:

        >>> samples = samples.reshape((-1,1)) # ndims = 1
        >>> samples = samples.reshape((1,-1)) # nsamples = 1

    Example:

        >>> # Turn 0:1 samples into -1:1
        >>> import numpy as np
        >>> norm_values = np.linspace(0,1,5).reshape((-1,1))
        >>> real_values = scale_samples(norm_values, [(-1,1)])
        >>> print(real_values)
        [[-1. ]
         [-0.5]
         [ 0. ]
         [ 0.5]
         [ 1. ]]
        >>> # Logarithmically scale to 1:10000
        >>> real_values = scale_samples(norm_values, [(1,1e4)] do_log=True)
        >>> print(real_values)
        [[  1.00000000e+00]
         [  1.00000000e+01]
         [  1.00000000e+02]
         [  1.00000000e+03]
         [  1.00000000e+04]]
    """
    norms = np.asarray(samples_norm)
    if len(norms.shape) != 2:
        raise ValueError()
    ndims = norms.shape[1]
    if not hasattr(do_log, "__iter__"):
        do_log = ndims * [do_log]
    logs = np.asarray(do_log)
    lims_norm = np.array([limits_norm for i in logs])
    _lims = [np.log10(limit) if log else limit for limit, log in zip(limits, logs)]
    lims = np.array(_lims)

    slopes = (lims[:, 1] - lims[:, 0]) / (lims_norm[:, 1] - lims_norm[:, 0])
    samples = slopes * (norms - lims_norm[:, 0]) + lims[:, 0]
    samples[:, logs] = pow(10, samples[:, logs])
    return samples


def process_scale(scale):
    if scale is not None:
        raw = ast.literal_eval(scale)
        processed = np.array(raw, dtype=float).tolist()
        return processed


class MakeSamples(CliCommand):
    def run(
        self,
        seed,
        n,
        dims,
        sample_type,
        scale,
        scale_factor,
        outfile,
        x0,
        x1,
        n_line,
        hard_bounds,
    ):
        np.random.seed(seed)
        n_samples = n
        n_dims = dims
        hard_bounds = hard_bounds
        sample_type = sample_type
        if sample_type == "random":
            x = np.random.random((n_samples, n_dims))
        elif sample_type == "grid":
            subdivision = int(pow(n_samples, 1 / float(n_dims)))
            temp = [np.linspace(0, 1.0, subdivision) for i in range(n_dims)]
            X = np.meshgrid(*temp)
            x = np.stack([xx.flatten() for xx in X], axis=1)
        elif sample_type == "lhs":
            x = doe.lhs(n_dims, samples=n_samples)
        elif sample_type == "lhd":
            _x = doe.lhs(n_dims, samples=n_samples)
            x = norm(loc=0.5, scale=0.125).ppf(_x)
        elif sample_type == "star":
            _x = doe.doe_star.star(n_dims)[0]
            x = 0.5 * (_x + 1.0)  # transform to center at 0.5 (range 0-1)
        elif sample_type == "ccf" or sample_type == "ccc" or sample_type == "cci":
            _x = np.unique(doe.ccdesign(n_dims, face=sample_type), axis=0)
            x = 0.5 * (_x + 1.0)
        else:
            raise ValueError(sample_type + " is not a valid choice for sample_type!")

        scales = process_scale(scale)

        if scales is not None:
            limits = []
            do_log = []
            for scale in scales:
                limits.append((scale[0], scale[1]))
                if len(scale) < 3:
                    scale.append("linear")
                if scale[2] == "log":
                    do_log.append(True)
                else:
                    do_log.append(False)
            x = scale_samples(x, limits, do_log=do_log)

        # scale the whole box
        x = scale_factor * x

        # add x0
        if x0 is not None:
            x0 = np.atleast_2d(np.load(x0))
            if scales is not None:
                sa = scale_factor * np.array(scales)[:, :2].astype("float")
                center = np.mean(sa, axis=1)
            else:
                center = scale_factor * 0.5
            # Loop over all x0 points
            all_x = []
            for _x0 in x0:

                _x = x + _x0 - center

                # replace the first entry with x0 for the random ones
                if sample_type == "lhs" or sample_type == "lhd":
                    _x[0] = _x0
                else:  # add it for the stencil points
                    _x = np.insert(_x, 0, _x0, axis=0)

                if x1 is not None:
                    x1 = np.load(x1)
                    line_range = np.linspace(0, 1, n_line + 1, endpoint=False)[1:]
                    line_samples = _x0 + np.outer(line_range, (x1 - _x0))
                    _x = np.vstack((_x, line_samples))
                all_x.append(_x)

            x = np.vstack(all_x)

        if hard_bounds:
            if scales is None:
                x = np.clip(x, 0, 1)
            else:
                for i, dim in enumerate(scales):
                    x[:, i] = np.clip(x[:, i], dim[0], dim[1])

        print(x)

        np.save(outfile, x)



import json


def process_args(args):
    result = []
    for path in args.instring.split("\n"):
        with open(path, "r") as json_file:
            result.append(json.load(json_file))

    with open(args.outfile, "w") as outfile:
        json.dump(result, outfile)



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-<PENDING>
# All rights reserved.
# This file is part of merlin-spellbook, Version: 0.4.2.
#
# For details, see https://github.com/LLNL/merlin-spellbook.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
For more info about conduit, see:
http://llnl-conduit.readthedocs.io/en/latest/user.html.

For documentation and examples, check out the conduit unit tests at:
https://github.com/LLNL/conduit/tree/master/src/tests
"""
import logging
import os

import numpy as np


LOG = logging.getLogger(__name__)

try:
    import conduit
    import conduit.relay.io
except:
    LOG.warning("conduit not found")


def determine_protocol(fname):
    """
    Determines a file protocol based on file name extension.
    """
    _, ext = os.path.splitext(fname)
    if ext.startswith("."):
        protocol = ext.lower().strip(".")
    else:
        raise ValueError(
            "{0} needs an ext (eg .hdf5) to determine protocol!".format(fname)
        )
    # Map .h5 to .hdf5
    if protocol == "h5":
        protocol = "hdf5"
    return protocol


def pack_conduit_node_from_dict(d):
    """
    If d is a dict, returns a conduit node, unpacked recursively using the
    dictionary to create the conduit hierarchy.

    If d is not a dict, simply returns d, to avoid extra conduit nodes being
    created. The intent is that the return value is then assigned to a conduit
    node by the caller and conduit will handle packing it appropriately.

    Conduit currently supports following basic types:
    int32, uint32, int64, uint64, float32, float64, numpy arrays of (uint32,
    uin64, float32, float64), strings.

    There is a bug in the current version of conduit that prevents usage of
    numpy arrays of signed ints.

    Notably, lists, dicts, tuples, and unicode strings can not be assigned
    directly as a conduit node. numpy is your friend. This method will auto
    convert tuples into numpy arrays of float64, will iterate over lists and
    give them each a separate entry in the hierarchy, labeled by their index,
    and assume any dicts are meant as part of the hierarchy description so the
    dict hierarchy ends up in the conduit hierarchy.
    """
    if isinstance(d, dict):
        node = conduit.Node()
        for k in d:
            try:
                node[str(k)] = pack_conduit_node_from_dict(d[k])
            except TypeError:
                try:
                    if isinstance(d[k], list):
                        node[str(k)] = create_conduit_node_from_list(d[k])
                    elif isinstance(d[k], tuple):
                        node[str(k)] = np.array(list(d[k]), dtype="float64")
                    else:
                        LOG.error("Conduit does not support following value", k, d[k])
                except TypeError:
                    LOG.error("Conduit does not support following value", k, d[k])
        return node
    else:
        if isinstance(d, (list, tuple, np.ndarray)):
            d_a = np.asarray(d)
            node = conduit.Node()
            node["data"] = d_a
            if len(d_a.shape) > 1:
                node["metadata/shape"] = d_a.shape
            return node
        if d is None:
            return "None"
        return d


def create_conduit_node_from_list(results, prefix=""):
    """
    Results must be a list of dictionaries or conduit nodes, each of which can
    have dictionaries, but all leaves must be a type that conduit can support
    (see description in pack_conduit_node_from_dict).
    """
    n = conduit.Node()
    for i in range(0, len(results)):
        try:
            n[str(prefix) + str(i)] = pack_conduit_node_from_dict(results[i])
        except TypeError:
            LOG.error("Unable to pack ", results[i])

    return n


def dump_node(
    conduit_node,
    fname,
    dump_options=(
        ("hdf5/chunking/threshold", 2000),
        ("hdf5/chunking/chunk_size", 2000),
    ),
):
    """
    Saves a conduit node to disk. Protocol determined by fname extension.
    Protocol can be conduit_bin, hdf5, json,
    silo, json64, and probably others. Will turn on compression for hdf5.
    """
    protocol = determine_protocol(fname)
    # If hdf5, turn on compression.
    if protocol == "hdf5":
        save_options = conduit.Node()
        for opt in dump_options:
            save_options[opt[0]] = opt[1]
        try:
            conduit.relay.io.save(conduit_node, fname, options=save_options)
        except TypeError:  # Conduit version needs to be updated.
            LOG.error(
                "Unable to customize save: please upgrade conduit to "
                "expose save options!"
            )
            conduit.relay.io.save(conduit_node, fname)
    else:
        conduit.relay.io.save(conduit_node, fname)


def load_node(fname, path="/"):
    """
    Read a conduit file and return a node with data under the path.
    If path is None, just returns the node handle.
    """
    handle = load_node_handle(fname, mode="r")
    if path is not None:
        if path == "/":
            n = conduit.Node()
            handle.read(n)
            handle.close()
            return n
        if handle.has_path(path):
            n = conduit.Node()
            handle.read(n, path)
            handle.close()
            return n
        else:
            print(f"ERROR in conduit_bundler.load_node: invalid_path {path}")
            handle.close()
            return None
    else:
        return handle


def load_node_handle(fname, mode="r"):
    """
    Read a conduit file node handle. Does not read into memory.
    """
    if os.path.exists(fname):
        options = conduit.Node()
        options["mode"] = mode
        handle = conduit.relay.io.IOHandle()
        handle.open(fname, options=options)
        return handle
    else:
        raise IOError("No such file: " + fname)



from types import SimpleNamespace

import click

from spellbook.utils import OptionEatAll


@click.command()
@click.option(
    "--output",
    required=False,
    default="output.json",
    type=str,
    help="output file",
)
@click.option(
    "--vars",
    required=False,
    default="results.hdf5",
    type=str,
    help="variables to write. specified as space separated name=VALUE",
    cls=OptionEatAll,
)
@click.option(
    "--splitter",
    required=False,
    default="/",
    type=str,
    help="key to indicate a nested value, default: /",
)
@click.option(
    "--verbose",
    is_flag=True,
    required=False,
    default=False,
    type=bool,
    help="print output, default False",
)
@click.option(
    "--indent",
    is_flag=True,
    required=False,
    default=False,
    type=bool,
    help="indent with new lines, default False",
)
def cli(output, vars, splitter, verbose, indent):
    """
    write a serialized file from cli arguments.
    """
    from spellbook.data_formatting import serialize

    args = SimpleNamespace(
        **{
            "output": output,
            "vars": vars,
            "splitter": splitter,
            "verbose": verbose,
            "indent": indent,
        }
    )
    serialize.parse_args(args)



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of merlin-spellbook.
#
# For details, see https://github.com/LLNL/merlin-spellbook and
# https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

import json
import sys

import numpy as np

import spellbook.ml.surrogates as surrogates


try:
    import cPickle as pickle
except ImportError:
    import pickle


def load_infile(args):
    with np.load(args.infile) as data:
        if args.X is not None:
            X = stack_arrays(data, args.X)  # inputs
        elif "X" in data.keys():
            X = data["X"]
        else:
            X = data[data.files[0]]

        if args.y is not None:
            y = stack_arrays(data, args.y)  # outputs
        elif "y" in data.keys():
            y = data["y"]
        else:
            y = data[data.files[1]]
    return X, y


def make_regressor(args):
    if args.reg_args is None:
        reg_args = {}
    else:
        reg_args = args.reg_args

    regr = surrogates.sklearnRegressors.factory(args.regressor, **reg_args)
    X, y = load_infile(args)

    n_samples_X = X.shape[0]
    n_samples_y = y.shape[0]

    if n_samples_X != n_samples_y:
        raise ValueError("n_samples_X != n_samples_y")

    # single feature or sample reshape as appropriate for sklearn
    if n_samples_y == 1:
        y.reshape((1, -1))
    elif len(y.shape) == 1:
        y.reshape((-1, 1))

    regr.fit(X, y)
    with open(args.outfile, "wb") as f:
        pickle.dump(regr, f)


def stack_arrays(data, delimited_names, delimiter=","):
    stacked = np.vstack([data[name] for name in delimited_names.split(delimiter)])
    return stacked.T



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of merlin-spellbook.
#
# For details, see https://github.com/LLNL/merlin-spellbook and
# https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

import numpy as np


try:
    import cPickle as pickle
except ImportError:
    import pickle


def predict(args):
    regr = pickle.load(open(args.reg, "rb"))

    X = np.load(args.infile)

    new_y = regr.predict(X)
    np.save(open(args.outfile, "wb"), new_y)



from types import SimpleNamespace

import click

from spellbook.utils import OptionEatAll


@click.command()
@click.option(
    "-infiles",
    required=True,
    type=str,
    cls=OptionEatAll,
    help="whitespace separated list of files to collect",
)
@click.option(
    "-outfile",
    required=False,
    default="results.hdf5",
    type=str,
    help="aggregated file root. If chunking will insert _n before extension",
)
@click.option(
    "-chunk_size",
    required=False,
    default=None,
    type=int,
    help="number of files to chunk together. Default (None): don't chunk",
)
@click.option(
    "-add_uuid",
    required=False,
    default=False,
    is_flag=True,
    help="auto insert a unique id for each file, creating a tree on the fly",
)
def cli(infiles, outfile, chunk_size, add_uuid):
    """
    Convert a list of conduit-readable files into a single big conduit node. Simple append, so nodes that already exist will get a name change to conflict-uuid
    """
    from spellbook.data_formatting.conduit.python import collector

    args = SimpleNamespace(
        **{
            "infiles": infiles,
            "outfile": outfile,
            "chunk_size": chunk_size,
            "add_uuid": add_uuid,
        }
    )
    collector.process_args(args)



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin Spellbook.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
Script for running command line interface tests.
Built for 1) manual use and 2) continuous integration.
"""
import argparse
import os
import shutil
import sys
import time
from contextlib import suppress
from glob import glob
from re import search
from subprocess import PIPE, Popen


OUTPUT_DIR = "cli_test_studies"


def run_single_test(name, test, test_label="", buffer_length=50):
    dot_length = buffer_length - len(name) - len(str(test_label))
    print(f"TEST {test_label}: {name}{'.'*dot_length}", end="")
    command = test[0]
    conditions = test[1]
    if not isinstance(conditions, list):
        conditions = [conditions]

    start_time = time.time()
    process = Popen(command, stdout=PIPE, stderr=PIPE, shell=True)
    stdout, stderr = process.communicate()
    end_time = time.time()
    total_time = end_time - start_time
    if stdout is not None:
        stdout = stdout.decode("utf-8")
    if stderr is not None:
        stderr = stderr.decode("utf-8")
    return_code = process.returncode

    info = {
        "total_time": total_time,
        "command": command,
        "stdout": stdout,
        "stderr": stderr,
        "return_code": return_code,
    }

    # ensure all test conditions are satisfied
    for condition in conditions:
        condition.ingest_info(info)
        passed = condition.passes
        if passed is False:
            break

    return passed, info


def clear_test_studies_dir():
    """
    Deletes the 'test_studies' directory, in order to preserve
    state each time cli tests are run.
    """
    with suppress(FileNotFoundError):
        shutil.rmtree(f"./{OUTPUT_DIR}")


def process_test_result(passed, info, is_verbose, exit):
    """
    Process and print test results to the console.
    """
    # if the environment does not contain necessary programs, exit early.
    if passed is False and "spellbook: command not found" in info["stderr"]:
        print(f"\nMissing from environment:\n\t{info['stderr']}")
        return None
    elif passed is False:
        print("FAIL")
        if exit is True:
            return None
    else:
        print("pass")

    if is_verbose is True:
        print(f"\tcommand: {info['command']}")
        print(f"\telapsed time: {round(info['total_time'], 2)} s")
        if info["return_code"] != 0:
            print(f"\treturn code: {info['return_code']}")
        if info["stderr"] != "":
            print(f"\tstderr:\n{info['stderr']}")

    return passed


def run_tests(args, tests):
    """
    Run all inputted tests.
    :param `tests`: a dictionary of
        {"test_name" : ("test_command", [conditions])}
    """
    selective = False
    n_to_run = len(tests)
    if args.ids is not None and len(args.ids) > 0:
        if not all(x > 0 for x in args.ids):
            raise ValueError(f"Test ids must be between 1 and {len(tests)}, inclusive.")
        selective = True
        n_to_run = len(args.ids)

    print(f"Running {n_to_run} integration tests...")
    start_time = time.time()

    total = 0
    failures = 0
    for test_name, test in tests.items():
        test_label = total + 1
        if selective and test_label not in args.ids:
            total += 1
            continue
        try:
            passed, info = run_single_test(test_name, test, test_label)
        except BaseException as e:
            print(e)
            passed = False
            info = None

        result = process_test_result(passed, info, args.verbose, args.exit)
        if result is None:
            print("Exiting early")
            return 1
        if result is False:
            failures += 1
        total += 1

    end_time = time.time()
    total_time = end_time - start_time

    if failures == 0:
        print(f"Done. {n_to_run} tests passed in {round(total_time, 2)} s.")
        return 0
    print(
        f"Done. {failures} tests out of {n_to_run} failed after {round(total_time, 2)} s.\n"
    )
    return 1


class Condition:
    def ingest_info(self, info):
        """
        This function allows child classes of Condition
        to take in data AFTER a test is run.
        """
        for key, val in info.items():
            setattr(self, key, val)

    @property
    def passes(self):
        print("Extend this class!")
        return False


class ReturnCodeCond(Condition):
    """
    A condition that some process must return 0
    as its return code.
    """

    def __init__(self, expected_code=0):
        """
        :param `expected_code`: the expected return code
        """
        self.expected_code = expected_code

    @property
    def passes(self):
        return self.return_code == self.expected_code


class NoStderrCond(Condition):
    """
    A condition that some process have an empty
    stderr string.
    """

    @property
    def passes(self):
        return self.stderr == ""


class RegexCond(Condition):
    """
    A condition that some body of text MUST match a
    given regular expression. Defaults to stdout.
    """

    def __init__(self, regex, negate=False):
        """
        :param `regex`: a string regex pattern
        """
        self.regex = regex
        self.negate = negate

    def is_within(self, text):
        """
        :param `text`: text in which to search for a regex match
        """
        return search(self.regex, text) is not None

    @property
    def passes(self):
        if self.negate:
            return not self.is_within(self.stdout)
        return self.is_within(self.stdout) or self.is_within(self.stderr)


def define_tests():
    """
    Returns a dictionary of tests, where the key
    is the test's name, and the value is a tuple
    of (shell command, condition(s) to satisfy).
    """

    # shortcut string variables
    config_dir = "./CLI_TEST_MERLIN_CONFIG"

    return {
        "spellbook": ("spellbook", ReturnCodeCond(1)),
        "spellbook help": ("spellbook --help", ReturnCodeCond()),
        "spellbook version": ("spellbook --version", ReturnCodeCond()),
        "spellbook serialize": (
            "spellbook serialize --output thing.json --verbose --vars top/middle/bottom=1 top/middle/bottom2=2 top/middle/string=spam top/bool=false top/float=1.3e-9 top/1/junk=nan ; rm thing.json",
            RegexCond(
                '{"top": {"1": {"junk": NaN}, "bool": true, "float": 1.3e-09, "middle": {"bottom": 1, "bottom2": 2, "string": "spam"}}}'
            ),
        ),
        "spellbook make-samples": (
            "spellbook make-samples ; rm samples.npy",
            ReturnCodeCond(),
        ),
        # "spellbook learn": ("spellbook learn", ReturnCodeCond()),
        # "spellbook predict": ("spellbook predict", ReturnCodeCond()),
        # "spellbook collect": ("spellbook collect", ReturnCodeCond()),
        #    etc... lots of tests for each of these subcommands are needed.
        #    Make your own Condition classes and insert them after the command string.
    }


def setup_argparse():
    parser = argparse.ArgumentParser(description="run_tests cli parser")
    parser.add_argument(
        "--exit",
        action="store_true",
        help="Flag for stopping all testing upon first failure",
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Flag for more detailed output messages"
    )
    parser.add_argument(
        "--ids",
        action="store",
        dest="ids",
        type=int,
        nargs="+",
        default=None,
        help="Provide space-delimited ids of tests you want to run."
        "Example: '--ids 1 5 8 13'",
    )
    return parser


def main():
    """
    High-level CLI test operations.
    """
    parser = setup_argparse()
    args = parser.parse_args()

    tests = define_tests()

    clear_test_studies_dir()
    result = run_tests(args, tests)
    clear_test_studies_dir()
    return result


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

import json
import os
import sys


""" Serializes a command-line input variable list"""


def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value


def maybe_numeric_or_bool(string):
    if string.lower() == "false" or string.lower() == "true":
        return bool(string)
    if string.isdigit():
        return int(string)
    try:
        return float(string)
    except:
        return string


def nested_dict(var_list, splitter="/"):
    output = {}
    for v in var_list:
        keys, val = v.split("=")
        keylist = keys.split(splitter)
        value = maybe_numeric_or_bool(val)
        nested_set(output, keylist, value)
    return output


def parse_args(args):
    output = nested_dict(args.vars, splitter=args.splitter)
    dumpargs = {"sort_keys": True}
    if args.indent:
        dumpargs["indent"] = 4
    if args.verbose:
        print(json.dumps(output, **dumpargs))
    with open(args.output, "w") as f:
        json.dump(output, f, **dumpargs)



from types import SimpleNamespace

import click


@click.command()
@click.option(
    "-instring",
    required=False,
    default="",
    type=str,
    help="whitespace separated list of files to collect",
)
@click.option(
    "-outfile",
    required=False,
    default="results.hdf5",
    type=str,
    help="output file",
)
def cli(instring, outfile):
    """
    Collect many json files into a single json file
    """
    from spellbook.data_formatting import collector

    args = SimpleNamespace(**{"instring": instring, "outfile": outfile})
    collector.process_args(args)



###############################################################################
# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin-Spellbook, Version: 0.4.2.
#
# For details, see https://github.com/LLNL/merlin-spellbook.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

import os

from setuptools import find_packages, setup


version = __import__("spellbook").__version__

extras = ["dev"]


def readme():
    with open("README.md") as f:
        return f.read()


# The reqs code from celery setup.py
def _strip_comments(l):
    return l.split("#", 1)[0].strip()


def _pip_requirement(req):
    if req.startswith("-r "):
        _, path = req.split()
        return reqs(*path.split(os.path.sep))
    return [req]


def _reqs(*f):
    return [
        _pip_requirement(r)
        for r in (
            _strip_comments(l)
            for l in open(os.path.join(os.getcwd(), "requirements", *f)).readlines()
        )
        if r
    ]


def reqs(*f):
    """Parse requirement file.
    Example:
        reqs('default.txt')          # requirements/default.txt
        reqs('extras', 'redis.txt')  # requirements/extras/redis.txt
    Returns:
        List[str]: list of requirements specified in the file.
    """
    trl = [req for subreq in _reqs(*f) for req in subreq]
    rl = [r for r in trl if "-e" not in r]
    return rl


def install_requires():
    """Get list of requirements required for installation."""
    return reqs("release.txt")


def extras_require():
    """Get map of all extra requirements."""
    return {x: reqs(x + ".txt") for x in extras}


setup(
    name="merlin-spellbook",
    author="Merlin Dev team",
    author_email="merlin@llnl.gov",
    version=version,
    description="The building blocks of workflows!",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="machine learning workflow utilities",
    url="https://github.com/LLNL/merlin-spellbook",
    license="MIT",
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=install_requires(),
    extras_require=extras_require(),
    entry_points={
        "console_scripts": [
            "spellbook=spellbook.main:main",
        ]
    },
    include_package_data=True,
    zip_safe=False,
)



from types import SimpleNamespace

import click


@click.command()
@click.option(
    "-infile",
    required=False,
    default="results.npz",
    type=str,
    help=".npz file with X and y data",
)
@click.option(
    "-X",
    required=False,
    default=None,
    type=str,
    help="variable(s) in infile for the input, defaults to X; can be a comma-delimited list",
)
@click.option(
    "-y",
    required=False,
    default=None,
    type=str,
    help="variable(s) in infile for the output, defaults to y; can be a comma-delimited list",
)
@click.option(
    "-outfile",
    required=False,
    default="random_forest_reg.pkl",
    type=str,
    help="file to pickle the regressor to",
)
def cli(infile, x, y, outfile):
    """
    Use sklearn to make a regressor
    """
    from spellbook.ml import learn_alt as learn

    args = SimpleNamespace(
        **{
            "infile": infile,
            "X": x,
            "y": y,
            "outfile": outfile,
        }
    )
    learn.random_forest(args)



import argparse
import json
import sys

import numpy as np


def setup_argparse():
    parser = argparse.ArgumentParser(description="Translate .json into numpy")

    parser.add_argument(
        "-input",
        help=".json file with X and y data in each sample",
        default="results.json",
    )
    parser.add_argument(
        "-output", help=".npz file with the arrays", default="results.npz"
    )
    parser.add_argument(
        "-schema", help="schema for a single sample data", default="features.json"
    )
    return parser


def process_args(args):
    samples = json.load(open(args.input, "r"))
    schema = json.load(open(args.schema, "r"))

    input_array_dict = {}
    output_array_dict = {}
    for s in samples:
        make_data_array_dict(input_array_dict, s["inputs"], schema["inputs"])
        make_data_array_dict(output_array_dict, s["outputs"], schema["outputs"])

    for path in output_array_dict:
        output_array_dict[path] = np.array(output_array_dict[path])

    for path in input_array_dict:
        input_array_dict[path] = np.array(input_array_dict[path])

    X = np.vstack([input_array_dict[x] for x in input_array_dict])

    output_array_dict["X"] = X.T

    np.savez(args.output, **output_array_dict)


def generate_scalar_path_pairs(node, schema, path=""):
    for child in node:
        # only process children that are schema compatible
        if child in schema.keys():
            if isinstance(node[child], dict):
                if isinstance(schema[child], dict):
                    for pair in generate_scalar_path_pairs(
                        node[child], schema[child], path=path + child + "/"
                    ):
                        yield pair
            else:
                if not isinstance(schema[child], dict):
                    yield path + child, node[child]


def make_data_array_dict(d, node, schema):
    for path, datum in generate_scalar_path_pairs(node, schema):
        if path in d:
            d[path].append(datum)
        else:
            d[path] = [datum]


def main():
    parser = setup_argparse()
    args = parser.parse_args()
    process_args(args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

import os
import shutil
import sys

import numpy as np


""" Merges npz files. Modified from https://jiafulow.github.io/blog/2019/02/17/merge-arrays-from-multiple-npz-files/"""


def find_max_dims(arrays):
    dims = np.atleast_2d(arrays[0]).shape
    for a in arrays[1:]:
        dims = np.max((dims, np.atleast_2d(a).shape), axis=0)
    return dims


def pad_many(arrays, dims, dont_pad_first=False, value=np.nan):
    fixed = []
    zeros = np.zeros_like(dims)
    for a in arrays:
        pad_dist = dims - np.atleast_2d(a).shape
        if dont_pad_first:
            pad_dist[0] = 0
        padder = np.column_stack((zeros, pad_dist))
        fixed.append(
            np.pad(np.atleast_2d(a), padder, mode="constant", constant_values=value)
        )
    return fixed


def stack_jagged(arrays, dont_pad_first=True, stack_func=np.vstack, **kwargs):
    dims = find_max_dims(arrays)
    padded = pad_many(arrays, dims, dont_pad_first, **kwargs)
    result = stack_func(padded)
    return result


class Stacker(object):
    def __init__(self):
        self.d = {}
        self.dout = {}

    def run(self, target, source, force=False):

        print("Target file: {0}".format(target))

        if not force:
            if os.path.isfile(target):
                print(
                    "stack_npz error opening target file (does {0} exist?).".format(
                        target
                    )
                )
                print('Pass "-f" argument to force re-creation of output file.')
                return

        n_source = len(source)
        if n_source == 1:
            print("Only one source file given! Not stacking, just doing a copy.")
            shutil.copy(source[0], target)
            print("DONE")
            return

        # Loop over the source files
        for i, s in enumerate(source):
            print("stack_npz Source file {0}: {1}".format(i, s))
            with np.load(s) as data:
                if i == 0:
                    for k in data.files:
                        self.d[k] = []
                # Loop over all the keys
                for k in data.files:
                    self.d[k].append(data[k])

        # Merge arrays via np.hstack()
        print("stacking...")
        for k, v in self.d.items():
            try:
                vv = stack_jagged(v)
                self.dout[k] = vv
            except:
                print(f"Error stacking {k}")

        # Write to the target file
        np.savez_compressed(target, **self.dout)
        print("DONE")


def process_args(args):
    stacker = Stacker()
    stacker.run(args.target, args.source, force=args.force)



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of merlin-spellbook.
#
# For details, see https://github.com/LLNL/merlin-spellbook and
# https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

from __future__ import print_function

from sklearn import base


try:
    from sklearn.utils import all_estimators
except ImportError:
    from sklearn.utils.testing import all_estimators


class sklearnRegressors(object):
    """Scikit learn regressor factory.

    Usage:
    import surrogates
    rf1 = surrogates.sklearnRegressors.factory('RandomForestRegressor', n_estimators=5, max_depth=3)
    gp1 = surrogates.sklearnRegressors.factory('GaussianProcessRegressor')
    choices = surrogates.sklearnRegressors.names()
    """

    def reg_dict():
        _all_regressors = {}
        estimators = all_estimators()
        for name, class_ in estimators:
            if issubclass(class_, base.RegressorMixin):
                _all_regressors[name] = class_
        return _all_regressors

    all_regs = reg_dict()

    @classmethod
    def factory(cls, name, *args, **kwargs):
        if name in cls.all_regs:
            return cls.all_regs[name](*args, **kwargs)
        else:
            raise ValueError(
                "Unknown regressor name "
                + name
                + "! For valid choices see sklearnRegressors.names()"
            )

    @classmethod
    def names(cls):
        return sorted(cls.all_regs.keys())


def test_factory():

    regressors = sklearnRegressors.names()
    for r in regressors:
        _ = sklearnRegressors.factory(r)


def test_random_forest():

    rf1 = sklearnRegressors.factory(
        "RandomForestRegressor", n_estimators=10, max_depth=5
    )
    rf2 = sklearnRegressors.factory(
        "RandomForestRegressor", n_estimators=2, max_depth=3
    )

    assert rf1.n_estimators == 10
    assert rf1.max_depth == 5

    assert rf2.n_estimators == 2
    assert rf2.max_depth == 3


if __name__ == "__main__":
    test_factory()
    test_random_forest()



import argparse
import sys

import numpy as np
from sklearn.ensemble import RandomForestRegressor


try:
    import cPickle as pickle
except ImportError:
    import pickle


FOREST_DEFAULTS = {"max_depth": 2, "random_state": 0, "n_estimators": 100}


def load_infile(args):
    with np.load(args.infile) as data:
        if args.X is not None:
            X = stack_arrays(data, args.X)  # inputs
        elif "X" in data.keys():
            X = data["X"]
        else:
            X = data[data.files[0]]

        if args.y is not None:
            y = stack_arrays(data, args.y)  # outputs
        elif "y" in data.keys():
            y = data["y"]
        else:
            y = data[data.files[1]]
    return X, y


def random_forest(args):
    forest_args = FOREST_DEFAULTS

    regr = RandomForestRegressor(**forest_args)
    X, y = load_infile(args)

    n_samples_X = X.shape[0]
    n_samples_y = y.shape[0]

    if n_samples_X != n_samples_y:
        raise ValueError("n_samples_X != n_samples_y")

    # single feature or sample reshape as appropriate for sklearn
    if n_samples_y == 1:
        y.reshape((1, -1))
    elif len(y.shape) == 1:
        y.reshape((-1, 1))

    regr.fit(X, y)
    with open(args.outfile, "wb") as f:
        pickle.dump(regr, f)


def stack_arrays(data, delimited_names, delimiter=","):
    stacked = np.vstack([data[name] for name in delimited_names.split(delimiter)])
    return stacked.T



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of merlin-spellbook.
#
# For details, see https://github.com/LLNL/merlin-spellbook and
# https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

import sys

import numpy as np

from spellbook.data_formatting.conduit.python import conduit_bundler as cb


WARN = ""
try:
    import conduit
    import conduit.relay.io
except:
    WARN = "\nWARNING: conduit not found."


def process_args(args):
    print(WARN)
    protocol = cb.determine_protocol(args.output)
    # Faster loader, just read metadata
    data_loader = cb.load_node_handle(args.input)
    first_data = conduit.Node()
    data_loader.read(first_data, data_loader.list_child_names()[0])
    if args.schema == "auto":
        schema_json = first_data.to_json()
    elif "," in args.schema:
        sub_list = args.schema.split(",")
        schema_node = conduit.Node()
        for item in sub_list:
            schema_node[item] = first_data[item]
        schema_json = schema_node.to_json()
    else:
        with open(args.schema, "r") as f:
            schema_json = f.read()

    g = conduit.Generator(schema_json, "json")
    schema = conduit.Node()
    g.walk_external(schema)

    data_paths = []
    for path, _ in generate_scalar_path_pairs(schema):
        data_paths.append(path)
    samples = data_loader.list_child_names()

    # Walk through all the samples and create a unified list (ie pack into a
    # dictionary of lists)
    all_dict = {}
    for s in samples:
        filtered_node = conduit.Node()
        for path in data_paths:
            sample_path = "/".join((s, path))
            if data_loader.has_path(sample_path):
                data_loader.read(filtered_node[path], sample_path)
            else:
                filtered_node[
                    sample_path
                ] = np.nan  # if a value is missing, that could be a problem
        make_data_array_dict(all_dict, filtered_node)

    for dat in all_dict.keys():
        all_dict[dat] = np.vstack(all_dict[dat])
    # Save according to output extension, either numpy or conduit-compatible
    if protocol == "npz":
        np.savez(args.output, **all_dict)
    else:
        n = cb.pack_conduit_node_from_dict(all_dict)
        cb.dump_node(n, args.output)


def generate_scalar_path_pairs(node, path=""):
    """ Walk through the node finding the paths to the data and the data """
    children = node.child_names()
    for child in children:
        if isinstance(node[child], conduit.Node):
            for pair in generate_scalar_path_pairs(
                node[child], path=path + child + "/"
            ):
                yield pair
        else:
            yield path + child, node[child]


def make_data_array_dict(d, node):
    """ Pact a node to the end of the list in the dictionary with same name path """
    for path, datum in generate_scalar_path_pairs(node):
        # patch for older versions of conduit
        datum = np.array(datum)
        if path in d:
            d[path].append(datum)
        else:
            d[path] = [datum]



import click


@click.command()
@click.option(
    "-seed",
    required=False,
    default=None,
    type=int,
    help="random number seed for generating samples",
)
@click.option(
    "-n",
    required=False,
    default=100,
    type=int,
    help="number of samples",
)
@click.option(
    "-dims",
    required=False,
    default=2,
    type=int,
    help="number of dimensions",
)
@click.option(
    "-sample_type",
    required=False,
    default="random",
    type=str,
    help="type of sampling. options: random, grid, lhs, lhd, star, ccf, ccc, cci. If grid, will try to get close to the correct number of samples. for lhd min-max correspond to +- 3 sigma range",
)
@click.option(
    "-scale",
    required=False,
    default=None,
    type=str,
    help='ranges to scale results in form "[(min,max,type),(min, max,type)]" where type = "linear" or "log" (optional: defaults to linear if omitted)',
)
@click.option(
    "-scale_factor",
    required=False,
    default=1.0,
    type=float,
    help="scale factor to appy to all ranges (stacks with -scale)",
)
@click.option(
    "-outfile",
    required=False,
    default="samples.npy",
    type=click.File("wb"),
    help="name of output .npy file",
)
@click.option(
    "-x0",
    required=False,
    default=None,
    type=click.File("rb"),
    help="file with optional point to center samples around, will be added as first entry",
)
@click.option(
    "-x1",
    required=False,
    default=None,
    type=click.File("rb"),
    help="file with x1 to add points between x0 and x1 (non inclusive) along a line",
)
@click.option(
    "-n_line",
    required=False,
    default=100,
    type=int,
    help="number of samples along a line between x0 and x1",
)
@click.option(
    "--hard-bounds",
    is_flag=True,
    required=False,
    default=False,
    type=bool,
    help="force all points to lie within -scale",
)
def cli(
    seed,
    n,
    dims,
    sample_type,
    scale,
    scale_factor,
    outfile,
    x0,
    x1,
    n_line,
    hard_bounds,
):
    """
    Generate some samples!
    """
    from spellbook.sampling import make_samples

    obj = make_samples.MakeSamples()
    obj.run(
        seed,
        n,
        dims,
        sample_type,
        scale,
        scale_factor,
        outfile,
        x0,
        x1,
        n_line,
        hard_bounds,
    )
